<?php 
// Software released under the General Public License (version 2 or later), available at
// http://www.gnu.org/copyleft/gpl.html

require_once "astronomy/solarsystem/SolarSystemConstants";

/****************************************************************************************
Computations related to moon phases (new / full moon, quarters).

@author   <a href="http://www.tig12.net">Thierry Graff</a>
@history 2003.01.25 : Creation from a split of astronomy.php
@history	: 2008.09.19 13:04 : Put in a PHP5 class 
****************************************************************************************/

class MoonNodes{
  
  //********************* getNodeDate ******************************
  /**
    Computes the date of passage of the moon through its node, for given julian day.
    From Meeus, Astronomical Algorithms, 2nd edition, chap. 51.
    @param $jd The julian day.
    @param $node The required moon node, using a constant of SolarSystemConstants.php.
    @param $beforeAfter A string indicating if the required phase should be searched
                   after or before $jd. Possible values : "before" or "after".
    @return The julian day of the desired event
    @history 2003.01.25 Creation from OCR (Optical Character Recognition) of Meeus' book.
  */
  public static function getNodeDate($node, $jd, $beforeAfter){
    $dateTime = jdToDate($jd);
    $year = getDecimalYear($dateTime);
    $k = ($year - 2000.05) * 13.4223;
    //
    $T = $k / 1342.23;
    $T2 = $T*$T;
    $T3 = $T2*$T;
    $T4 = $T3*$T;
    //
    $D  = deg2rad(mod360(183.6380 + 331.73735682*$k + 0.0014852*$T2
        + 0.00000209*$T3 - 0.000000010*$T4));
    $M  = deg2rad(mod360(17.4006 + 26.82037250*$k + 0.0001186*$T2 + 0.00000006*$T3));
    $Mp = deg2rad(mod360(38.3776 + 355.52747313*$k + 0.0123499*$T2
        + 0.000014627*$T3 - 0.000000069*$T4));
    $om = deg2rad(mod360(123.9767 - 1.44098956*$k + 0.0020608*$T2
        + 0.00000214*$T3 - 0.000000016*$T4));
    $V  = deg2rad(mod360(299.75 + 132.85*$T - 0.009173*$T2));
    $P  = deg2rad(mod360($om + 272.75 - 2.3*$T));
    $E  = 1 - 0.002516*$T - 0.0000074*$T2; // formula 47.6
    //
    $Dx2 = 2*$D;
    $Dx4 = 4*$D;
    $JDE = 2451565.1619 + 27.212220817*$k
         + 0.0002762*$T2
         + 0.000000021*$T3
         - 0.000000000088*$T4
         - 0.4721*sin($Mp)
         - 0.1649*sin($Dx2)
         - 0.0868*sin($Dx2 - $Mp)
         + 0.0084*sin($Dx2 + $Mp)
         - 0.0083*sin($Dx2 - $M) * $E
         - 0.0039*sin($Dx2 - $M - $Mp) * $E
         + 0.0034*sin(2*$Mp)
         - 0.0031*sin($Dx2 - 2*$Mp)
         + 0.0030*sin($Dx2 + $M) * $E
         + 0.0028*sin($M - $Mp) * $E
         + 0.0026*sin($M) * $E
         + 0.0025*sin($Dx4)
         + 0.0024*sin($D)
         + 0.0022*sin($M + $Mp) * $E
         + 0.0017*sin($om)
         + 0.0014*sin($Dx4 - $Mp)
         + 0.0005*sin($Dx2 + $M - $Mp) * $E
         + 0.0004*sin($Dx2 - $M + $Mp) * $E
         - 0.0003*sin($Dx2 - 2*$M) * $E
         + 0.0003*sin($Dx4 - $M) * $E
         + 0.0003*sin($V)
         + 0.0003*sin($P);
    return $JDE;
  }// end getNodeDate
  
  
}// end class
?>