<?php 
// Software released under the General Public License (version 2 or later), available at
// http://www.gnu.org/copyleft/gpl.html
/********************************************************************************
An object of this class represents a date.
Once created, a date object can't be modified.
It holds the following private variables, accessible via __get() method :
    year,
    month,
    day,
    hour,
    min,
    sec,
    jd (julian day)
    dayOfYear (day of year),
    decimalYear (year expressed as a decimal number)
    decimalDay (day expressed as a decimal number)
    dtu (timezone offset, in minutes)

Limited support for internationalized output.

@licence GPL
@author   <a href="http://www.tig12.net">Thierry Graff</a>
@history 2008.08.31 17:25 : Creation
@todo bug in __get('dayOfWeek')
****************************************************************************************/

class Date{
  
  // basic infos
  /** Year : integer */
  private $year = null;
  /** Month (1-12) : integer */
  private $month = null;
  /** Day (1-31) : integer*/
  private $day = null;
  /** Hour (0-23) : integer */
  private $hour = null;
  /** Minuts (0-59) :integer */
  private $minute = null;
  /** Seconds (0-59) : decimal */
  private $second = null;
  /** Timezone offset, in minuts */
  private $dtu = null;
  
  // precision
  /** Precision, in seconds */
  private $precision = null;
  /** Number used to specify the precision */
  private $precision_nb = null;
  /** Unit used to specify the precision */
  private $precision_unit = null;
  
  // Redundant infos
  /** Julian day : decimal */
  private $jd = null;
  /** Day of year (0-366) : integer */
  private $dayOfYear = null;
  /** Day of week, between 0 (sunday) and 6 (saturday) */
  private $dayOfWeek = null;
  /**
    Year, expressed as a decimal number.
    Fractional part corresponds to the the number of days, h, m, s elapsed since the beginning of the year
  */
  private $decimalYear = null;
  /**
    Day, expressed as a decimal number.
    Fractional part corresponds to the the number of h, m, s elapsed since the beginning of the day
  */
  private $decimalDay = null;
  
  //=================================================================================
  //                                              Constructor
  //=================================================================================
  
  /** 
    Constructs a new Date object.
    @param $when To specify the date / time represented by this object.
                  Can be in one of the following format :
                  - A number (integer or decimal) : considered to be a julian day
                  - A string : 
                    - 'now' corresponds to the current instant ; 
                    - any other string is interpreted as a date expressed in <a href='todo'>jetheme format</a>
    @throws Exception if $when doesn't represent a valid date.
  */
  public function __construct($when='now'){
    if(is_numeric($when)){
      $this->jd = $when;
      // todo : compute precision fields
      //$this->precision
      //$this->precision_nb
      //$this->precision_unit
    }
    else if($when == 'now'){
      $dt = new DateTime('now');
      $this->year = $dt->format('Y');
      $this->month = $dt->format('m');
      $this->day = $dt->format('d');
      $this->hour = $dt->format('H');
      $this->minute = $dt->format('i');
      $this->second = $dt->format('s');
      // todo : handle dtu
    }
    else{
      // string in jetheme format
      // Isolate the precision
      $pos = strPos($when, 'P');
      if($pos !== false){
        $precision = substr($when, $pos+1);
        $when2 = substr($when, 0, $pos);
        // check precision format
        $pattern = '/(\d+)(\D{1,2})/';
        preg_match($pattern, $precision, $matches);
        if(count($matches) == 0)
          throw new Exception("Error in date <b>$when</b> : Precision uncorrectly formatted : '$precision'");
        switch(strToUpper($matches[2])){
          case 'S' :  $precision2 = $precision; break;
          case 'M' :  $precision2 = $precision * 60; break;
          case 'H' :  $precision2 = $precision * 3600; break;
          case 'D' :  $precision2 = $precision * 86400; break;
          case 'W' :  $precision2 = $precision * 604800; break; // 86400 * 7
          case 'MO' :  $precision2 = $precision * 2592000; break; // 86400 * 30
          case 'Y' :  $precision2 = $precision * 31557600; break; // 86400 * 365.25
          case 'C' :  $precision2 = $precision * 3155760000; break;// 86400 * 365.25 * 100
          default : throw new Exception("Error in date <b>$when</b> : Invalid character to specify precision unit : '{$matches[2]}'");
        }
        $this->precision = $precision2;
        $this->precision_nb = $matches[1];
        $this->precision_unit = $matches[2];
        $when = $when2; // $when2 was useful only to have original $when in error messages
      }
      // Compute the other fields
      $parsed = date_parse($when);
      $this->year = $parsed['year'];
      $this->month = $parsed['month'];
      $this->day = $parsed['day'];
      $this->hour = $parsed['hour'];
      $this->minute = $parsed['minute'];
      $this->second = $parsed['second'] + $parsed['fraction'];
      if(isset($parsed['zone'])) $this->dtu = $parsed['zone'];
    }
    if(is_null($this->precision)){
      // todo : compute the precision from the smallest time unit specified
    }
  }// end __construct
  
  
  //=================================================================================
  //                                              Member methods
  //=================================================================================
  
  //**************************** __get *************************************
  /** getter for private fields */
  public function __get($fieldName){
//echo "<br/>in __get, fieldName = $fieldName";
    if($fieldName == 'jd'){
      if(is_null($this->jd)) $this->jd = self::dateToJd($this->year, $this->month, $this->day, $this->hour, $this->minute, $this->second);
      return $this->jd;
    }
    // precision fields always computed in constructor
    else if($fieldName == 'precision'){
      return $this->precision;
    }
    else if($fieldName == 'precision_nb'){
      return $this->precision_nb;
    }
    else if($fieldName == 'precision_unit'){
      return $this->precision_unit;
    }
    else{
      if(is_null($this->year)){
        // year = null => the date was built with jd
        list($this->year, $this->month, $this->day, $this->hour, $this->minute, $this->second) = self::jdToDate($this->jd);
      }
      switch($fieldName){
        case 'year' : return $this->year; break;
        case 'month' : return $this->month; break;
        case 'day' : return $this->day; break;
        case 'hour' : return $this->hour; break;
        case 'minute' : return $this->minute; break;
        case 'second' : return $this->second; break;
        case 'dayOfYear' :
          if(is_null($this->dayOfYear)) $this->dayOfYear = self::dayOfYear($this->year, $this->month, $this->day);
          return $this->dayOfYear;
        break;
        case 'dayOfWeek' :
          if(is_null($this->dayOfWeek)) $this->dayOfWeek = self::dayOfWeek($this->jd);
          return $this->dayOfWeek;
        break;
        case 'decimalYear' :
          if(is_null($this->decimalYear)) $this->decimalYear = self::decimalYear($this->year, $this->month, $this->day, $this->hour, $this->minute, $this->second);
          return $this->decimalYear;
        break;
        case 'decimalHour' :
          if(is_null($this->decimalHour)) $this->decimalHour = self::decimalHour($this->hour, $this->minute, $this->second);
          return $this->decimalHour;
        break;
      }// end switch
    }// end else
  }// end __get
  
  
  //=================================================================================
  //                                              Static methods
  //=================================================================================
  
  
  //****************** dateToJd **************
  /**
    Calculates the number of julian days elapsed since jan 0.5 4712 BC, from a date and an hour.
    Formula from Meeus2, p 61
  */
  public static function dateToJd($year, $month, $day, $hour, $minute, $second){
    $decDay = $day + $hour/24 + $minute/1440 + $second/86400;
    if($month>2){
      $y1 = $year;
      $m1 = $month;
    }
    else{
      $y1 = $year-1;
      $m1 = $month+12;
    }
    //
    if(($year+$month/100.0 + $decDay/10000.0) >= 1582.1015){
      $a = floor($y1/100.0);
      $b = 2-$a+floor($a/4.0);
    }
    else{
      $b = 0;
    }
    $jd = floor(365.25*($y1+4716)) + floor(30.6001*($m1+1)) + $decDay + $b - 1524.5;
    return $jd;
  }// end dateToJd
  
  
  //****************** jdToDate ****************************
  /**
    Computes a date from a julian day (the number of julian days elapsed since jan 0.5 4712 BC).
    <br/>Source : Jean Meeus, "Astronomical Algorithms", 2nd edition, chap. 7.
    @param $jd The julian day.
    @return An array containing 6 elements : years, months, days, hours, minutes, decimal seconds.
  */
  public static function jdToDate($jd){
    $DAYS_PER_YEAR = 365.25;
    if($jd < 0)
      throw new Exception("jdToDate() not valid for negative julian days");
    $z = floor($jd + 0.5);
    $f = $jd + 0.5 - $z;
    if($z < 2299161.0)
      $a = $z;
    else{
      $alpha = floor(($z - 1867216.25) / 36524.25);
      $a = $z + 1 + $alpha - floor($alpha/4.0);
    }
    $b = $a + 1524.0;
    $c = floor(($b - 122.1) / $DAYS_PER_YEAR);
    $d = floor($DAYS_PER_YEAR * $c);
    $e = floor(($b - $d) / 30.6001);
    // months
    if ($e < 14.0) $res[1] = $e - 1;
    else           $res[1] = $e - 13;
    // years
    if ($res[1] > 2) $res[0] = $c - 4716;
    else             $res[0] = $c - 4715;
    // days
    $res[2] = $b - $d - floor(30.6001 * $e);
    // hours
    $hours = 24 * $f; // decimal hours
    $res[3] = floor($hours);
    // minuts
    $minutes = ($hours - floor($hours)) * 60.0; // decimal minuts
    $res[4] = floor($minutes);
    // seconds
    $secs = ($minutes - floor($minutes)) * 60.0; // decimal seconds
    $res[5] = floor(1000 * $secs) / 1000; // round to 1/1000 second
    return $res;
  }// end jdToDate
  
  
  //****************** decimalDay ****************************
  /**
    Returns the decimal part of the day a time represents.
    Inverse of {@link decimalDayToHms()} 
  */
  public static function decimalDay($hour, $min, $sec){
    $decMin = $min + $sec/60;
    $decHour = $hour + $decMin/60;
    return $decHour/24;
  }// end decimalDay
  
  
  //****************** decimalDayToHms ****************************
  /**
    Computes hours, minutes, seconds from a decimal part of the day.
    Inverse of {@link decimalDay()} 
    Ex : <code>decimalDayToHMS(13.5) = decimalDayToHMS(0.5) = array(12, 0, 0)</code>
    @param Decimal $decDay A day number expressed as a decimal.
    @return Array containing 3 elements : hours, minutes, seconds
    @todo unit test
  */
  public static function decimalDayToHms($decDay){
    $decDay = $decDay - floor($decDay); // remove the integer part
    $hours = floor($decDay * 24);
    $rest = $decDay - $hours / 24;
    $minutes = floor($rest * 1440); // 1440 = 24 * 60
    $rest = $rest - $minutes / 1440;
    $seconds = floor($rest * 86400); // 86400 = 24 *60 * 60
    return array($hours, $minutes, $seconds);
  }// end decimalDayToHms
  
  
  //****************** decimalYear ****************************
  /**
    Computes a date as a decimal year.
  */
  public static function decimalYear($year, $month, $day, $hour=0, $min=0, $sec=0){
    if(self::isLeapYear($year)) $n = 366;
    else $n = 365;
    $formattedDate = uFormatStrings::addZeroes($year, 4). "-" . uFormatStrings::addZeroes($month) . "-" . uFormatStrings::addZeroes($day);
    $dayOfYear = self::dayOfYear($year, $month, $day) - 1;
    $decDay = self::decimalDay($hour, $min, $sec);
    return $year + ($dayOfYear + $decDay)/$n;
  }// end decimalYear
  
  
  // ************************ dayOfYear ********************
  /**
    Returns the day of year, from 1 to 365 or 366.
    Formula from Meeus2, p. 65.
  */
  public static function dayOfYear($year, $month, $day){
    $K = ( self::isLeapYear($year) ? 1 : 2);
    return floor(275 * $month / 9) - $K * floor(($month + 9) / 12) + $day - 30;
  }// end dayOfYear
  
  
  // ************************ dayOfWeek ********************
  /**
    Returns the day of week, from 0 (sunday) to 6 (saturday).
    Formula from Meeus2, p. 65.
    @param $jd Julian day. Should correspond to a date et 0h UT (otherwise may return a wrong value)
  */
  public static function dayOfWeek($jd){
echo "<br/>jd : $jd";
    $jd += 1.5;
    return $jd % 7;
  }// end dayOfWeek
  
  
  //**************************** isLeapYear *************************************
  /**
    Tests if a year is leap (bissextile).
    @return true if <code>$year</code> is positive and corresponds to a leap year.
  */
  public static function isLeapYear($year){
    if ($year < 0) return false;
    if($year < 1582){
      // Julian calendar
      if ($year % 4 == 0)
        return true;
      return false;
    }
    // Gregorian calendar
    if ( ($year % 4 == 0 && $year % 100 != 0) || ($year % 400 == 0) )
      return true;
    return false;
  }// end isLeapYear
  
  
  //=================================================================================
  //                                              Code related to output
  //=================================================================================
  
  /** Date output format. JJ/MM/AAAA (fr) or MM/DD/YYYY (en) */
  const DATE_FORMAT_1 = 'date_format_1';
  
  /** Date output format. ex : 27 janv. 2008 (fr) or jan 27, 2008 (en) */
  const DATE_FORMAT_2 = 'date_format_2';
  
  /** Date output format. Ex :   janvier 2008 (fr) or January 27, 2008 (en) */
  const DATE_FORMAT_3 = 'date_format_3';
  
  /** Date output format. samedi 27 janvier 2008 (fr) or Saturday january 27, 2008 (en) */
  const DATE_FORMAT_4 = 'date_format_4';
  
  /** Hour output format. HHhMM (ex : 3h02 ; 12h34) */
  const HOUR_FORMAT_1 = 'hour_format_1';
  
  /** Hour output format. HH:MM:SS (ex : 03:02:23) */
  const HOUR_FORMAT_2 = 'hour_format_2';
  
  // *** English strings ***
  
  /** Array containing english month names */
  public static $MONTHS_FULL_EN = array(
    'january', 'febuary', 'march', 'april', 'may', 'june', 'july', 'august', 'september', 'october', 'november', 'december'
  );
  
  /** Array containing english abbreviated month names */
  public static $MONTHS_ABBR_EN = array(
    'jan', 'feb', 'mar', 'apr', 'may', 'jun', 'jul', 'aug', 'sep', 'oct', 'nov', 'dec'
  );
  
  /** Array containing english days of week */
  public static $WEEKDAYS_EN = array(
    'sunday', 'monday', 'tuesday', 'wednesday', 'thursday', 'friday', 'saturday'
  );
  
  // *** French strings ***
  
  /** Array containing french month names */
  public static $MONTHS_FULL_FR = array(
    'janvier', 'février', 'mars', 'avril', 'mai', 'juin', 'juillet', 'août', 'septembre', 'octobre', 'novembre', 'décembre'
  );
  
  /** Array containing french abbreviated month names */
  public static $MONTHS_ABBR_FR = array(
    'janv.', 'fév.', 'mars', 'avr.', 'mai', 'juin', 'juil.', 'août', 'sept.', 'oct.', 'nov.', 'déc.'
  );
  
  /** Array containing french days of week */
  public static $WEEKDAYS_FR = array(
    'dimanche', 'lundi', 'mardi', 'mercredi', 'jeudi', 'vendredi', 'samedi'
  );
  
  //**************************** toString *************************************
  /**
    Returns a string representing this date.
    @param $options Associative array of key / value pairs ; possible keys :
                  - 'lang' : ISO 639 code of language. 'en' and 'fr' are currently supported
                  Default : value of G::$user_language
                  - 'ucFirst' : Boolean ; should the first character of the returned string be upper case?
                  Default : true
                  - 'date_format' : use DATE_FORMAT_x constants of this class
                  Default : DATE_FORMAT_3
                  - 'year_conditional' : Boolean ; if false the year is always part of the returned string 
                  if false, the year is part of the returned string only if different from current year
                  Default : false
                  - 'date_hour_separator' : String ; when hour also returned, separator used between date and hour
                  Default : ' '
                  - 'hour_output' : Boolean ; should the hour be part of the returned string ?
                  Default : true
                  - 'hour_format' : use HOUR_FORMAT_x constants of this class
                  Default : HOUR_FORMAT_2
  */
  public function toString($options = array()){
    // parameter stuff
    if(!isset($options['lang'])) $options['lang'] = G::get('user_language');
    if(!isset($options['date_format'])) $options['date_format'] = self::DATE_FORMAT_3;
    if(!isset($options['year_conditional'])) $options['year_conditional'] = false;
    if(!isset($options['hour_output'])) $options['hour_output'] = true;
    if(!isset($options['hour_format'])) $options['hour_format'] = self::HOUR_FORMAT_2;
    if(!isset($options['ucFirst'])) $options['ucFirst'] = true;
    if(!isset($options['date_hour_separator'])) $options['date_hour_separator'] = ' ';
    //
    $strRes = '';
    switch($options['date_format']){
      case self::DATE_FORMAT_1 : 
        $strRes .= $this->toString_dateFormat1($options['lang'], $options['year_conditional']);
      break;
      case self::DATE_FORMAT_2 : 
        $strRes .= $this->toString_dateFormat2($options['lang'], $options['year_conditional']);
      break;
      case self::DATE_FORMAT_3 : 
        $strRes .= $this->toString_dateFormat3($options['lang'], $options['year_conditional']);
      break;
      case self::DATE_FORMAT_4 : 
        $strRes .= $this->toString_dateFormat4($options['lang'], $options['year_conditional']);
      break;
    }
    if($options['ucFirst']){
      $strRes = ucFirst($strRes);
    }
    if($options['hour_output'] === false){
      return $strRes;
    }
    // hour
    $strRes .= $options['date_hour_separator'];
    switch($options['hour_format']){
      case self::HOUR_FORMAT_1 : 
        $strRes .= $this->toString_hourFormat1();
      break;
      case self::HOUR_FORMAT_2 : 
        $strRes .= $this->toString_hourFormat2();
      break;
    }
    return $strRes;
  }// end toString
  
  
  // *************************** toString_dateFormat1 ***************************
  /**
    Auxiliary method of {@link toString()}
    @param $lang See {@link toString()}
    @param $year_conditional See parameter $options['year_conditional'] of {@link toString()}
  */
  private function toString_dateFormat1($lang, $year_conditional){
    $res = '';
    if($lang == 'en'){
      $res .= $this->__get('month') . '/' . $this->__get('day');
    }
    else if($lang == 'fr'){
      $res .= $this->__get('day') . '/' . $this->__get('month');
    }
    if($year_conditional && $this->__get('year') == date('Y')){
      return $res;
    }
    // year
    $res .= '/' . $this->__get('year');
    return $res;
  }// end toString_dateFormat1
  
  
  // *************************** toString_dateFormat2 ***************************
  /**
    Auxiliary method of {@link toString()}
    @param $lang See {@link toString()}
    @param $year_conditional See parameter $options['year_conditional'] of {@link toString()}
  */
  private function toString_dateFormat2($lang, $year_conditional){
    $res = '';
    if($lang == 'en'){
      $res .= self::$MONTHS_ABBR_EN[$this->__get('month') - 1] . ' ' . $this->__get('day');
    }
    else if($lang == 'fr'){
      $res .= $this->__get('day') . ' ' . self::$MONTHS_ABBR_FR[$this->__get('month') - 1];
    }
    if($year_conditional && $this->__get('year') == date('Y')){
      return $res;
    }
    // year
    if($lang == 'en'){
      $res .= ', ';
    }
    else if($lang == 'fr'){
      $res .= ' ';
    }
    $res .= $this->__get('year');
    return $res;
  }// end toString_dateFormat2
  
  
  // *************************** toString_dateFormat3 ***************************
  /**
    Auxiliary method of {@link toString()}
    @param $lang See {@link toString()}
    @param $year_conditional See parameter $options['year_conditional'] of {@link toString()}
  */
  private function toString_dateFormat3($lang, $year_conditional){
//echo "<br/>dateformat3 - month : " . $this->__get('month');
    $res = '';
    if($lang == 'en'){
      $res .= self::$MONTHS_FULL_EN[$this->__get('month') - 1] . ' ' . $this->__get('day');
    }
    else if($lang == 'fr'){
      $res .= $this->__get('day') . ' ' . self::$MONTHS_FULL_FR[$this->__get('month') - 1];
    }
    if($year_conditional && $this->__get('year') == date('Y')){
      return $res;
    }
    // year
    if($lang == 'en'){
      $res .= ', ';
    }
    else if($lang == 'fr'){
      $res .= ' ';
    }
    $res .= $this->__get('year');
    return $res;
  }// end toString_dateFormat3
  
  
  // *************************** toString_dateFormat4 ***************************
  /**
    Auxiliary method of {@link toString()}
    @param $lang See {@link toString()}
    @param $year_conditional See parameter $options['year_conditional'] of {@link toString()}
  */
  private function toString_dateFormat4($lang, $year_conditional){
    $res = '';
echo "<br/>dow : " . $this->dayOfWeek;
    if($lang == 'en'){
      $res .= self::$WEEKDAYS_EN[$this->__get('dayOfWeek')];
    }
    else if($lang == 'fr'){
      $res .= self::$WEEKDAYS_FR[$this->__get('dayOfWeek')];
    }
    return $res . ' ' . $this->toString_dateFormat3($lang, $year_conditional);
  }// end toString_dateFormat4
  
  
  // *************************** toString_hourFormat1 ***************************
  /**
    Auxiliary method of {@link toString()}
  */
  private function toString_hourFormat1(){
    return $this->__get('hour') . 'h' . uFormatStrings::addZeroes($this->__get('minute'));
  }// end toString_hourFormat1
  
  
  // *************************** toString_hourFormat2 ***************************
  /**
    Auxiliary method of {@link toString()}
  */
  private function toString_hourFormat2(){
    return uFormatStrings::addZeroes($this->__get('hour'))
       . ':' . uFormatStrings::addZeroes($this->__get('minute'))
       . ':' . uFormatStrings::addZeroes(round($this->__get('second')));
  }// end toString_hourFormat2
  
  
}// end class
 ?>